#!/usr/bin/env python3

import os
import sys
from pathlib import Path
import logging
import subprocess
from sepi_point.mutation_finder import MutationFinder
from sepi_point.seqtools import WgsData
import argparse
import re
import sepi_point.sepi_point as sp
from importlib import resources


def parse_args(argv):
    parser = argparse.ArgumentParser(description='Run sepi_point on a single isolate')
    parser.add_argument("-r", "--read_dir",
                        help = "Folder with paired end read files (fastq or fastq.gz)",
                        type=Path,
                        required = False)
    parser.add_argument("-a", "--assembly_dir",
                        help = "Folder with genome assemblies (fasta)",
                        type=Path,
                        required = False)
    parser.add_argument("-o", "--output",
                        help = "Output directory.",
                        type=Path,
                        required = True)
    parser.add_argument("-n", "--no_clean",
                        help = "Do not clean up sam and bam files after variant calling. Default False.",
                        action= "store_true",
                        default=False)
    args = parser.parse_args()
    return args





def main_cli():
    args = parse_args(argv=sys.argv)
    
    ### Check if necessary inputs are provided and that folders exist
    if args.read_dir is not None:
        if args.read_dir.exists():
            print(f"Checking for read files in {args.read_dir}")
        else:
            print(f"Provided read input folder {args.read_dir} does not exist. Exitting.")
            sys.exit()
    if args.assembly_dir is not None:
        if args.assembly_dir.exists():
            print(f"Checking for assembly files in {args.assembly_dir}")
        else:
            print(f"Provided assembly input folder {args.assembly_dir} does not exist. Exitting.")
            sys.exit()
    elif args.read_dir is None:
        print("No inputs provided. Please provide an input folder with paired end reads (-r) and/or input folder with assemblies (-a). Exitting.")
        sys.exit()
    
    ### Set up output directory
    try:
        args.output.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        print(f"Failed to setup output directory at {args.output}: {e}. Exitting.")
        sys.exit()

    log_file = args.output.joinpath("sepi_point.log")
    logger = sp.setup_logger(log_file=log_file)
    mutation_db_tsv = resources.files("sepi_point").joinpath("db").joinpath("mutations.tsv")
    mutation_db_fasta = resources.files("sepi_point").joinpath("db").joinpath("sequences.fasta")

    ###
    wgs_data = WgsData.from_folders(assembly_data_folder=args.assembly_dir, paired_end_read_data_folder=args.read_dir)
    print(f"Loaded data for {len(wgs_data)} samples with read and/or assembly data.")
    mf = MutationFinder()
    mf.load_and_check_db(mutation_db_tsv=mutation_db_tsv,sequence_db_fasta=mutation_db_fasta)

    all_sample_mutations= {}
    for sample_name, file_paths, info in wgs_data:
        if file_paths["r1_file"] is not None and file_paths["r2_file"] is not None:
            if file_paths["assembly_file"] is not None:
                output_name_fasta = f"{sample_name}_fasta"
                output_name_fastq = f"{sample_name}_fastq"
                output_dir_fasta = args.output.joinpath(output_name_fasta)
                output_dir_fasta.mkdir(exist_ok=True)
                logger.info(f"#####################################################################")
                logger.info(f"Running sepi_point on {sample_name}")
                logger.info(f"Using input {file_paths['assembly_file']}")
                logger.info(f"Printing output to {output_dir_fasta}")
                print(f"Running sepi_point on {sample_name}")
                snp_file = sp.run_nucmer_and_showsnps(assembly_file=file_paths["assembly_file"],
                                                   output_dir=output_dir_fasta, output_prefix=sample_name,
                                                   reference_fasta=mutation_db_fasta, logger=logger)
                sample_mutations = mf.get_mutations_from_nucmer_snps(nucmer_snp_file=snp_file)
                sample_mutation_summary = mf.summarize_sample_mutations(sample_mutations=sample_mutations).copy()
                mf.print_sample_mutations(mutation_summary=sample_mutation_summary,summary_output_file=output_dir_fasta.joinpath("results.tsv"))
                all_sample_mutations[output_name_fasta] = sample_mutation_summary
            else:
                output_name_fastq = sample_name
            output_dir_fastq = args.output.joinpath(output_name_fastq)
            output_dir_fastq.mkdir(exist_ok=True)
            logger.info(f"#####################################################################")
            logger.info(f"Running sepi_point on {sample_name}")
            logger.info(f"Using inputs {file_paths['r1_file']}, {file_paths['r2_file']}")
            logger.info(f"Printing output to {output_dir_fastq}")
            print(f"Running sepi_point on {sample_name}")
            vcf = sp.run_mapping_and_variant_calling(r1_file=file_paths["r1_file"], r2_file=file_paths["r2_file"],
                                                     output_dir=output_dir_fastq, output_prefix=sample_name,
                                                     reference_fasta=mutation_db_fasta,
                                                     no_clean=args.no_clean, logger=logger)
            sample_mutations = mf.get_mutations_from_vcf(vcf_file=vcf)
            sample_mutation_summary = mf.summarize_sample_mutations(sample_mutations=sample_mutations).copy()
            mf.print_sample_mutations(mutation_summary=sample_mutation_summary,summary_output_file=output_dir_fastq.joinpath("results.tsv"))
            all_sample_mutations[output_name_fastq] = sample_mutation_summary
        elif file_paths["assembly_file"] is not None:
            output_name_fasta = sample_name
            output_dir_fasta = args.output.joinpath(output_name_fasta)
            output_dir_fasta.mkdir(exist_ok=True)
            logger.info(f"#####################################################################")
            logger.info(f"Running sepi_point on {sample_name}")
            logger.info(f"Using input {file_paths['assembly_file']}")
            logger.info(f"Printing output to {output_dir_fasta}")
            print(f"Running sepi_point on {sample_name}")
            snp_file = sp.run_nucmer_and_showsnps(assembly_file=file_paths["assembly_file"],
                                                  output_dir=output_dir_fasta, output_prefix=sample_name,
                                                  reference_fasta=mutation_db_fasta, logger=logger)
            sample_mutations = mf.get_mutations_from_nucmer_snps(nucmer_snp_file=snp_file)
            sample_mutation_summary = mf.summarize_sample_mutations(sample_mutations=sample_mutations).copy()
            mf.print_sample_mutations(mutation_summary=sample_mutation_summary,summary_output_file=output_dir_fasta.joinpath("results.tsv"))
            all_sample_mutations[output_name_fasta] = sample_mutation_summary
            
    
    mf.print_sample_mutations_batch(mutation_summaries=all_sample_mutations,
                                    summary_output_file=args.output.joinpath("results.tsv"), matrix_output_file=args.output.joinpath("results.matrix.tsv"))
    

if __name__ == "__main__":
    main_cli()


import os
import sys
from sepi_point.seqtools import DnaSeq, ProteinSeq, NucleotideFasta, ProteinFasta, translate_dna
from pathlib import Path
import logging
import subprocess


class MutationFinder:

    def __init__(self):
        return(None)

    def load_and_check_db(self, mutation_db_tsv: Path, sequence_db_fasta: Path) -> None:
        self.tsv_path = os.path.abspath(mutation_db_tsv)
        self.fasta_path = os.path.abspath(sequence_db_fasta)
        self.sequences = NucleotideFasta.from_file(sequence_db_fasta)
        self.protein_sequences = self.sequences.translate()
        aa_mutation_dict = {}
        aa_mutation_list = []
        codon_mutation_dict = {}
        nt_mutation_dict = {}
        aa_to_codon = setup_aa_to_codon_table()
        with open(self.tsv_path) as f:
            firstline = True
            for line in f:
                line = line.rstrip('\n').split('\t')
                if firstline:
                    firstline = False
                    gene_idx = line.index("gene")
                    type_idx = line.index("type")
                    mutation_idx = line.index("mutation")
                    category_idx = line.index("category")
                else:
                    gene = line[gene_idx]
                    mutation = line[mutation_idx]
                    position = int(mutation[1:-1])
                    if line[type_idx] == "protein":
                        ref_aa = mutation[0]
                        if not self.protein_sequences[gene][0].sequence[position-1] == ref_aa:
                            print(f"Warning. Mutation tsv file contains mutation {gene}::{mutation}, but reference aa does not match fasta file")
                        alt_aa = mutation[-1]
                        alt_codons = aa_to_codon[alt_aa]
                        aa_mutation_list.append(gene+"::"+mutation)
                        if not gene in aa_mutation_dict:
                            aa_mutation_dict[gene] = {}
                            aa_mutation_dict[gene][str(position)] = [mutation]
                            codon_mutation_dict[gene] = {}
                            codon_mutation_dict[gene][str(position)] = alt_codons
                        elif not str(position) in aa_mutation_dict[gene]:
                            aa_mutation_dict[gene][str(position)] = [mutation]
                            codon_mutation_dict[gene][str(position)] = alt_codons
                        else:
                            aa_mutation_dict[gene][str(position)].append(mutation)
                            codon_mutation_dict[gene][str(position)] += alt_codons
                    else:
                        ref_nt = mutation[0]
                        if not self.sequences[gene][0].sequence[position-1] == ref_nt:
                            print(f"Warning. Mutation tsv file contains mutation {gene}::{mutation}, but reference nt does not match fasta file")
                        alt_nt = mutation[-1]
                        if not gene in nt_mutation_dict:
                            nt_mutation_dict[gene] = {}
                            nt_mutation_dict[gene][str(position)] = [alt_nt]
                        elif not str(position) in nt_mutation_dict[gene]:
                            nt_mutation_dict[gene][str(position)] = [alt_nt]
                        else:
                            nt_mutation_dict[gene][str(position)].append(alt_nt)

        self.aa_mutation_list = aa_mutation_list
        self.aa_mutation_dict = aa_mutation_dict
        self.codon_mutation_dict = codon_mutation_dict
        self.nt_mutation_dict = nt_mutation_dict
        return(None)

    def get_mutations_from_vcf(self, vcf_file: Path) -> dict:
        sample_mutation_dict = {}
        for gene in self.aa_mutation_dict:
            sample_mutation_dict[gene] = {}
        for gene in self.nt_mutation_dict:
            sample_mutation_dict[gene] = {}
        with open(vcf_file) as f:
            for line in f:
                if not line.startswith("##"):
                    if line.startswith("#"):
                        line = line.rstrip("\n").lstrip("#").split("\t")
                        CHROM_index = line.index("CHROM")
                        POS_index = line.index("POS")
                        REF_index = line.index("REF")
                        ALT_index = line.index("ALT")
                        FORMAT_index = line.index("FORMAT")
                        INFO_index = FORMAT_index + 1
                    else:
                        line = line.rstrip("\n").split("\t")
                        format = line[FORMAT_index].split(":")
                        info = line[INFO_index].split(":")
                        GT_index = format.index("GT")
                        if info[GT_index] == "1":
                            gene = line[CHROM_index]
                            pos = line[POS_index]
                            alt = line[ALT_index]
                            sample_mutation_dict[gene][pos] = alt
        return(sample_mutation_dict)

    def get_mutations_from_nucmer_snps(self, nucmer_snp_file: Path):
        sample_mutation_dict = {}
        for gene in self.nt_mutation_dict:
            sample_mutation_dict[gene] = {}
        with open(nucmer_snp_file) as f:
            for line in f:
                line = line.rstrip('\n').split('\t')
                gene = line[10]
                pos = line[0]
                alt = line[2]
                if not gene in sample_mutation_dict:
                    sample_mutation_dict[gene] = {}
                sample_mutation_dict[gene][pos] = alt
        return(sample_mutation_dict)

    def summarize_sample_mutations(self, sample_mutations: dict) -> dict:
        mutation_summary = {}
        for gene, alt_dict in self.nt_mutation_dict.items():
            for nt_position, alt_nts in alt_dict.items():
                if nt_position in sample_mutations[gene]:
                    for alt_nt in alt_nts:
                        if alt_nt in sample_mutations[gene][nt_position]:
                            ref_nt = self.sequences[gene][0].sequence[int(nt_position)-1]
                            nt_mut = ref_nt+nt_position+alt_nt
                            mutation_summary[gene+"::"+nt_mut] = [gene,nt_position,ref_nt,alt_nt,"",""]
        for gene, codon_dict in self.codon_mutation_dict.items():
            for aa_position, codons in codon_dict.items():
                start_position = int(aa_position)*3-2
                sample_codon = ""
                for position in range(start_position, start_position+3):
                    if gene in sample_mutations and str(position) in sample_mutations[gene]:
                        sample_codon += sample_mutations[gene][str(position)]
                    else:
                        sample_codon += self.sequences[gene][0].sequence[position-1]
                if sample_codon in codons:
                    ref_codon = self.sequences[gene][0].sequence[start_position-1:start_position+2]
                    ref_aa = translate_dna(ref_codon)
                    alt_aa = translate_dna(sample_codon)
                    aa_mut = ref_aa+aa_position+alt_aa
                    mutation_summary[gene+"::"+aa_mut] = [gene,aa_position,translate_dna(ref_codon),translate_dna(sample_codon),ref_codon,sample_codon]

        return(mutation_summary)

    @staticmethod
    def print_sample_mutations(mutation_summary: dict, summary_output_file: Path = None) -> None:
        print_header = ["Mutation","Gene","Position","Ref","Alt","Ref_codon","Alt_codon"]
        if summary_output_file is None:
            print("\t".join(print_header))
            for mutation, details in mutation_summary.items():
                printlist = [mutation]+details
                print("\t".join(printlist))
        else:
            o = open(summary_output_file,'w')
            o.write("\t".join(print_header)+"\n")
            for mutation, details in mutation_summary.items():
                printlist = [mutation]+details
                o.write("\t".join(printlist)+"\n")
            o.close()
        return(None)
    

    def print_sample_mutations_batch(self, mutation_summaries: dict[dict], summary_output_file: Path = None, matrix_output_file: Path = None) -> None:
        print_header = ["Sample","Mutation","Gene","Position","Ref","Alt","Ref_codon","Alt_codon"]
        matrix_header = ["Sample"]+self.aa_mutation_list
        o = open(summary_output_file,'w')
        o.write("\t".join(print_header)+"\n")
        om = open(matrix_output_file, 'w')
        om.write("\t".join(matrix_header)+"\n")
        for sample_name, mutation_summary in mutation_summaries.items():
            # write to long format output
            for mutation, details in mutation_summary.items():
                printlist = [sample_name, mutation]+details
                o.write("\t".join(printlist)+"\n")
            # write to presence/absence matrix output
            matrix_printlist = [sample_name]
            for mutation in self.aa_mutation_list:
                if mutation in mutation_summary:
                    matrix_printlist.append("1")
                else:
                    matrix_printlist.append("0")
            om.write("\t".join(matrix_printlist)+"\n")
        o.close()
        om.close()
        return(None)

    def iter_db_codons(self):
        for gene, codon_dict in self.codon_mutation_dict.items():
            for aa_position, codons in codon_dict.items():
                yield(gene, aa_position, codons)


def setup_aa_to_codon_table() -> dict:
    table = { 
            'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M', 
            'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T', 
            'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K', 
            'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R', 
            'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L', 
            'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P', 
            'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q', 
            'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R', 
            'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V', 
            'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A', 
            'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E', 
            'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G', 
            'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S', 
            'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L', 
            'TAC':'Y', 'TAT':'Y', 'TAA':'*', 'TAG':'*', 
            'TGC':'C', 'TGT':'C', 'TGA':'*', 'TGG':'W', 
            '---': '-'
        }
    aa_to_codon_table = {}
    for codon, aa in table.items():
        if aa in aa_to_codon_table:
            aa_to_codon_table[aa].append(codon)
        else:
            aa_to_codon_table[aa] = [codon]
    return(aa_to_codon_table)




#!/usr/bin/env python3

import os
import sys
from pathlib import Path
import logging
import subprocess
from sepi_point.mutation_finder import MutationFinder
import argparse
import re
from importlib import resources


def parse_args(argv):
    parser = argparse.ArgumentParser(description='Run sepi_point on a single isolate')
    parser.add_argument("-1", "--r1_file",
                        help = "Forward Illumina read file (fastq or fastq.gz)",
                        type=Path,
                        required = False)
    parser.add_argument("-2", "--r2_file",
                        help = "Reverse Illumina read file (fastq or fastq.gz)",
                        type=Path,
                        required = False)
    parser.add_argument("-a", "--assembly",
                        help = "Genome assembly (fasta)",
                        type=Path,
                        required = False)
    parser.add_argument("-o", "--output",
                        help = "Output directory.",
                        type=Path,
                        required = True)
    parser.add_argument("-s", "--sample_name",
                        help = "Sample name (will be auto-detected from input files if not supplied)",
                        type=str,
                        required = False)
    parser.add_argument("-n", "--no_clean",
                        help = "Do not clean up sam and bam files after variant calling. Default False.",
                        action= "store_true",
                        default=False)
    args = parser.parse_args()
    return args



def execute_cmd_and_log(cmd, logger, log_stdout=True, log_stderr=True) -> tuple[str, str]:
    logger.info(f"Running command: {cmd}")
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        encoding="utf-8",
    )
    stdout, stderr = process.communicate()
    if log_stdout and stdout and stdout is not None:
        logger.info(f"Shell command STDOUT: {stdout}")
    if log_stderr and stderr and stderr is not None:
        logger.error(f"Shell command STDERR: {stderr}")
    return stdout, stderr

def setup_logger(log_file, log_level="INFO") -> logging.RootLogger:
    logger = logging.getLogger()
    logging.basicConfig(
        level=log_level,
        filename=str(log_file),
        encoding="utf-8",
        filemode="w",
        format="{asctime} - {levelname} - {message}",
        style="{",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    return logger


def check_fastq_inputs(r1_file: Path, r2_file: Path, mutation_db_tsv: Path, mutation_db_fasta: Path, logger):
    """"
    Check if all inputs and database files are present for paired end read input
    """
    all_files_found = True
    if not r1_file.exists():
        all_files_found = False
        warning = f"No R1 input file found at {r1_file}"
        print(warning)
        logger.error(warning)
    if not r2_file.exists():
        all_files_found = False
        warning = f"No R2 input file found at {r2_file}"
        print(warning)
        logger.error(warning)
    if not mutation_db_fasta.exists():
        all_files_found = False
        warning = f"No mutation db tsv found at {mutation_db_tsv}"
        print(warning)
        logger.error(warning)
    if not mutation_db_fasta.exists():
        all_files_found = False
        warning = f"No mutation db fasta file found at {mutation_db_fasta}"
        print(warning)
        logger.error(warning)
    return(all_files_found)


def check_fasta_inputs(assembly_file: Path, mutation_db_tsv: Path, mutation_db_fasta: Path, logger):
    """"
    Check if all inputs and database files are present for assembly input
    """
    all_files_found = True
    if not assembly_file.exists():
        all_files_found = False
        warning = f"No assembly file found at {assembly_file}"
        print(warning)
        logger.error(warning)
    if not mutation_db_fasta.exists():
        all_files_found = False
        warning = f"No mutation db tsv found at {mutation_db_tsv}"
        print(warning)
        logger.error(warning)
    if not mutation_db_fasta.exists():
        all_files_found = False
        warning = f"No mutation db fasta file found at {mutation_db_fasta}"
        print(warning)
        logger.error(warning)
    return(all_files_found)


def run_mapping_and_variant_calling(r1_file: Path, r2_file: Path,
                                    output_dir: Path, output_prefix: str, 
                                    reference_fasta: Path,
                                    no_clean: bool, logger) -> Path:
    """"
    Run read-mapping and variant calling on paired end read files.
    Return path to vcf file
    """

    prefix = output_dir.joinpath(output_prefix)
    sam = Path(f"{prefix}.sam")                                   # sam file produced from bwa mem mapping to 
    bam = Path(f"{prefix}.bam")                                   # bam file filtered on q30 and only including mapped reads from primary alignments
    sorted_bam = Path(f"{prefix}.sorted.bam")                     # sorted bam file
    sorted_bam_idx = Path(f"{prefix}.sorted.bam.bai")             # sorted bam index file
    sorted_sam = Path(f"{prefix}.sorted.sam")                     # sorted sam file for parsing
    vcf = Path(f"{prefix}.vcf")                                   # variant calls

    if not vcf.exists() and not sorted_sam.exists() and not sorted_bam.exists() and not bam.exists() and  not sam.exists():
        cmd = f"bwa mem -v 1 -o {sam} {reference_fasta} {r1_file} {r2_file} 2> /dev/null"
        stdout, stderr = execute_cmd_and_log(cmd=cmd, logger=logger)
    else:
        logger.info(f"Sam file found at {sam}. Skipping bwa mem read mapping.")

    # Run samtools view to filter unmapped reads and convert to bam
    # -F 260 to only include mapped and exclude secondary alignments,
    if not vcf.exists() and not sorted_sam.exists() and not sorted_bam.exists() and not bam.exists():
        cmd = f"samtools view -q 30 -h -F 260 -O BAM -o {bam} {sam}"
        stdout, stderr = execute_cmd_and_log(cmd=cmd, logger=logger)
    else:
        logger.info(f"Bam file found at {bam}.")

    # Sort and index bam with samtools
    if not vcf.exists() and not sorted_sam.exists() and (not sorted_bam.exists() or not sorted_bam_idx.exists()):
        cmd = f"samtools sort {bam} -o {sorted_bam}; samtools index -o {sorted_bam_idx} {sorted_bam}"
        stdout, stderr = execute_cmd_and_log(cmd=cmd, logger=logger)
    else:
        logger.info(f"Sorted bam file found at {sorted_bam}.")


    # Run samtools view to convert bam to sam
    if not vcf.exists() and not sorted_sam.exists():
        cmd = f"samtools view -h -O SAM -o {sorted_sam} {sorted_bam}"
        stdout, stderr = execute_cmd_and_log(cmd=cmd, logger=logger)
    else:
        logger.info(f"Sorted sam file found at{sorted_sam}.")
    
    # run bcftools call to generate vcf
    if not vcf.exists():
        cmd = f"bcftools mpileup -A -f {reference_fasta} {sorted_sam} | bcftools call --ploidy 1 -mv -Ov -o {vcf}"
        stdout, stderr = execute_cmd_and_log(cmd=cmd, logger=logger,log_stdout=False, log_stderr=False)
    else:
        logger.info(f"Vcf file found at {vcf}.")
    
    if not no_clean:
        sam.unlink()
        bam.unlink()
        sorted_bam.unlink()
        sorted_bam_idx.unlink()
        sorted_sam.unlink()
        logger.info(f"Cleaned up bam and sam files from output_folder")

    return(vcf)


def run_nucmer_and_showsnps(assembly_file: Path, 
                            output_dir: Path, output_prefix: str,
                            reference_fasta: Path, logger) -> Path:
    """"
    Run nucmer and show-snps on assembled genome
    Return path to .snps file.
    """
    prefix = output_dir.joinpath(output_prefix)
    delta_file = Path(f"{prefix}.delta")
    snps_file = Path(f"{prefix}.snps")
    cmd = f"nucmer --minmatch 15 --prefix={prefix} {reference_fasta} {assembly_file}; show-snps -HrT {delta_file} > {snps_file}"
    if not snps_file.exists():
        stdout, stderr = execute_cmd_and_log(cmd=cmd, logger=logger,log_stdout=False, log_stderr=False)
    else:
        logger.info(f"Nucmer snps file found at {snps_file}.")
    return(snps_file)


def run_on_reads(r1_file: Path, r2_file: Path, output_dir: Path, output_prefix: str,
                 mutation_db_tsv: Path, mutation_db_fasta: Path, no_clean:bool, logger):
    all_files_found = check_fastq_inputs(r1_file=r1_file, r2_file=r2_file, mutation_db_tsv=mutation_db_tsv, mutation_db_fasta=mutation_db_fasta, logger=logger)
    if all_files_found:
        vcf = run_mapping_and_variant_calling(r1_file=r1_file, r2_file=r2_file,
                                        output_dir=output_dir, output_prefix=output_prefix,
                                        reference_fasta=mutation_db_fasta,
                                        no_clean=no_clean, logger=logger)
        mf = MutationFinder()
        mf.load_and_check_db(mutation_db_tsv=mutation_db_tsv,sequence_db_fasta=mutation_db_fasta)
        sample_mutations = mf.get_mutations_from_vcf(vcf_file=vcf)
    else:
        print(f"Input or reference DB files not found. Exiting.")
        sys.exit()
    return(mf, sample_mutations)

def run_on_assembly(assembly_file: Path, output_dir: Path, output_prefix: str,
                 mutation_db_tsv: Path, mutation_db_fasta: Path, logger):
    all_files_found = check_fasta_inputs(assembly_file=assembly_file, mutation_db_tsv=mutation_db_tsv, mutation_db_fasta=mutation_db_fasta,logger=logger)
    if all_files_found:
        snp_file = run_nucmer_and_showsnps(assembly_file=assembly_file,
                                            output_dir=output_dir, output_prefix=output_prefix,
                                            reference_fasta=mutation_db_fasta, logger=logger)
        mf = MutationFinder()
        mf.load_and_check_db(mutation_db_tsv=mutation_db_tsv,sequence_db_fasta=mutation_db_fasta)
        sample_mutations = mf.get_mutations_from_nucmer_snps(nucmer_snp_file=snp_file)
    else:
        sys.exit()
    return(mf, sample_mutations)

def main_cli():
    args = parse_args(argv=sys.argv)

    ### Set up output directory
    try:
        args.output.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        print(f"Failed to setup output directory at {args.output}: {e}")
        sys.exit()

    ### Check if input files are provided and set sample name if not provided
    if args.r1_file and args.r2_file:
        if not args.sample_name:
            re_match = re.match(r"(?P<sample_name>.+?)(?P<sample_number>(_S[0-9]+)?)(?P<lane>(_L[0-9]+)?)[\._]R?(?P<paired_read_number>[1|2])(?P<set_number>(_[0-9]+)?)(?P<file_extension>\.fastq\.gz)", args.r1_file.name)
            args.sample_name = re_match.group("sample_name")
    elif args.assembly:
        if not args.sample_name:
            args.sample_name = args.assembly.stem
    else:
        print("Error: Input files must be provided. Either an assembled genome with -a option or paired end reads with -1 and -2 options. See epi_point.py -h for more info.")
        sys.exit()
    log_file = Path(args.output).joinpath(f"{args.sample_name}.log")
    results_file = Path(args.output).joinpath(f"{args.sample_name}.results.tsv")

    logger = setup_logger(log_file=log_file)
    mutation_db_tsv = resources.files("sepi_point").joinpath("db").joinpath("mutations.tsv")
    mutation_db_fasta = resources.files("sepi_point").joinpath("db").joinpath("sequences.fasta")

    logger.info(f"#### RUNNING SEPI_POINT ####")
    logger.info(f"Checking for mutations found in {mutation_db_tsv}")
    logger.info(f"Reference fasta file: {mutation_db_fasta}")
    if args.r1_file and args.r2_file:
        logger.info(f"Using inputs {args.r1_file}, {args.r2_file}")
        mf, sample_mutations = run_on_reads(r1_file=args.r1_file, r2_file=args.r2_file,
                                            output_dir=args.output, output_prefix=args.sample_name,
                                            mutation_db_tsv=mutation_db_tsv, mutation_db_fasta=mutation_db_fasta,
                                            no_clean=args.no_clean, logger=logger)
            
    elif args.assembly:
        logger.info(f"Using input {args.assembly}")
        mf, sample_mutations = run_on_assembly(assembly_file=args.assembly,
                                               output_dir=args.output, output_prefix=args.sample_name, 
                                               mutation_db_tsv=mutation_db_tsv, mutation_db_fasta=mutation_db_fasta,
                                               logger=logger)
    else:
        logger.error("Input files must be provided. Either an assembled genome with -a option or paired end reads with -1 and -2 options.")
        print("Error: Input files must be provided. Either an assembled genome with -a option or paired end reads with -1 and -2 options. See epi_point.py -h for more info.")
    
    sample_mutation_summary = mf.summarize_sample_mutations(sample_mutations=sample_mutations)
    mf.print_sample_mutations(mutation_summary=sample_mutation_summary,summary_output_file=results_file)


if __name__ == "__main__":
    main_cli()
    



#!/usr/bin/env python3

import sys
from pathlib import Path
from sepi_point.mutation_finder import MutationFinder
from sepi_point.seqtools import WgsData
import argparse
from importlib import resources


def parse_args(argv):
    parser = argparse.ArgumentParser(description='Summarize sepi_point results')
    parser.add_argument("-r", "--results_dir",
                        help = "Folder with output folders from sepi_point runs.",
                        type=Path,
                        required = True)
    parser.add_argument("-o", "--output",
                        help = "Output Folder. Default <results_dir>",
                        type=Path,
                        required = False)
    args = parser.parse_args()
    return args


def main_cli():
    args = parse_args(argv=sys.argv)
    mutation_db_tsv = resources.files("sepi_point").joinpath("db").joinpath("mutations.tsv")
    mutation_db_fasta = resources.files("sepi_point").joinpath("db").joinpath("sequences.fasta")

    ###
    mf = MutationFinder()
    mf.load_and_check_db(mutation_db_tsv=mutation_db_tsv,sequence_db_fasta=mutation_db_fasta)
    all_sample_mutations = {}
    for folder in args.results_dir.iterdir():
        if folder.is_dir():
            sample_name = folder.name
            snps_file = folder.joinpath(f"{sample_name}.snps")
            vcf_file = folder.joinpath(f"{sample_name}.vcf")
            if vcf_file.exists():
                sample_mutations = mf.get_mutations_from_vcf(vcf_file=vcf_file)
                sample_mutation_summary = mf.summarize_sample_mutations(sample_mutations=sample_mutations).copy()
                all_sample_mutations[sample_name] = sample_mutation_summary
            elif snps_file.exists():
                sample_mutations = mf.get_mutations_from_nucmer_snps(nucmer_snp_file=snps_file)
                sample_mutation_summary = mf.summarize_sample_mutations(sample_mutations=sample_mutations).copy()
                all_sample_mutations[sample_name] = sample_mutation_summary
    print(f"Summarizing sepi_point results from {len(all_sample_mutations)} samples")
    if args.output:
        summary_output_file = args.output.joinpath("results.tsv")
        matrix_output_file = args.output.joinpath("results.matrix.tsv")
        if not args.output.exists():
            args.output.mkdir()
    else:
        summary_output_file = args.results_dir.joinpath("results.tsv")
        matrix_output_file = args.results_dir.joinpath("results.matrix.tsv")
    mf.print_sample_mutations_batch(mutation_summaries=all_sample_mutations,
                                    summary_output_file=summary_output_file, matrix_output_file=matrix_output_file)
    


if __name__ == "__main__":
    main_cli()


import os
import sys
from pathlib import Path
import pandas
import re
import numpy

class BioSeq:
    def __init__(self, name: str,sequence: str) -> None:
        self.name = name
        self.sequence = sequence.upper()

    def __len__(self):
        return(len(self.sequence))

    def __iter__(self):
        for nt in self.sequence:
            yield nt
        
    
    def __repr__(self):
        return f"< Biological sequence object {self.name} of length {len(self)} bp >"
    
    def __add__(self,other):
        return(BioSeq(self.name,self.sequence+other.sequence))

    @classmethod
    def from_fasta(cls, input_file: Path):
        startflag = True
        with open(input_file) as f:
            for line in f:
                line = line.rstrip('\n')
                if line[0] == ">":
                    if startflag:
                        seq = ""
                        startflag = False
                    else:
                        print(f"Fasta file {input_file} contains more than one entry. Only first entry is loaded into DnaSeq object.\nTo load all entries into Fasta object use Fasta.from_file({input_file})\n", file = sys.stderr)
                        f.close()
                        return(cls(sequence_name, seq))
                    sequence_name = line[1:]

                else:
                    seq += line
        f.close()
        return cls(sequence_name, seq)
    

    
    def print_fasta(self,file: Path,linelength: int = None) -> None:
        printlines = []
        if linelength is None:
            printlines += [">"+self.name,self.sequence]
        else:
            printlines += [">"+self.name] +  [(self.sequence[i:i+linelength]) for i in range(0, len(self.sequence), linelength)]
        
        o = open(file,"w")
        o.write("\n".join(printlines)+"\n")
        o.close
        return None
    


class DnaSeq(BioSeq):
    def __init__(self, name: str,sequence: str, phred_scores: str = None) -> None:
        self.name = name
        self.sequence = sequence.upper()
        self.phred_scores = phred_scores
    
    
    def __repr__(self):
        return f"< DNA sequence object {self.name} of length {len(self)} bp >"
    
    def reverse_complement(self):
        complement = {'A': 'T', 'T': 'A',
                      'C': 'G', 'G': 'C',
                      'R': 'Y', 'Y': 'R',
                      'S': 'W', 'W': 'S',
                      'K': 'M', 'M': 'K',
                      'B': 'V', 'V': 'B',
                      'D': 'H', 'H': 'D',
                      'N': 'N', 'X': 'X',
                      '-': '-', '.': '.'}
        reverse_complement = "".join(complement.get(base, base) for base in self.sequence[::-1])
        illegal_characters = set(self.sequence)-set(complement.keys())
        if len(illegal_characters) > 0:
            print("Non standard UIPAC characters found in sequence. Replacing with N when reverse complementing.", file=sys.stderr)
            for c in illegal_characters:
                reverse_complement = reverse_complement.replace(c,"N")
        if self.phred_scores is None:
            return(DnaSeq(self.name,reverse_complement))
        else:
            return(DnaSeq(self.name,reverse_complement,self.phred_scores[::-1]))
    
    def translate(self):
        seq = self.sequence.replace("-","")
        table = { 
            'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M', 
            'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T', 
            'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K', 
            'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R', 
            'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L', 
            'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P', 
            'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q', 
            'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R', 
            'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V', 
            'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A', 
            'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E', 
            'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G', 
            'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S', 
            'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L', 
            'TAC':'Y', 'TAT':'Y', 'TAA':'*', 'TAG':'*', 
            'TGC':'C', 'TGT':'C', 'TGA':'*', 'TGG':'W', 
        } 
        protein_seq = "" 
        mod = len(seq)%3
        #if not mod == 0:
        #    print(f"Warning: number of nucleotides in sequence is not divisible by 3. Ignoring {mod} nt at the end of sequence when translating to protein.", file=sys.stderr)
        illegal_codons = 0
        for i in range(0, len(seq)-mod, 3):
            codon = seq[i:i + 3]
            if len(set(codon)-{"A","C","G","T"}) > 0:
                illegal_codons += 1
                protein_seq += "X"
            else:
                protein_seq+= table[codon]
        if illegal_codons > 0:
            print(f"WARNING: {illegal_codons} codons containing non A/C/G/T found in {self.name}. X has been inserted on those positions.", file=sys.stderr)
        return(ProteinSeq(self.name,protein_seq))
    
    
    def __add__(self,other):
        if self.phred_scores is None or other.phred_scores is None:
            return(DnaSeq(self.name,self.sequence+other.sequence))
        else:
            return(DnaSeq(self.name,self.sequence+other.sequence,self.phred_scores,other.phred_scores))

    @property
    def iupac_counts(self):
        base_types = ["A","C","G","T"]
        amb_types = ["R","Y","S","W","K","M","B","D","H","M"]
        gap_types = [".","-"]
        nt_counts = {"A": 0,"C": 0,"G": 0,"T": 0, "N": 0, 
                     "R": 0,"Y": 0,"S": 0,"W": 0,"K": 0,"M": 0,"B": 0,"D": 0,"H": 0,"M": 0,
                     ".": 0,"-": 0,
                     "ACGT": 0, "amb": 0, "N": 0, "gap": 0, "nonIUPAC": 0}
        for nt in self.sequence:
            try:
                nt_counts[nt] += 1
            except KeyError:
                nt_counts["nonIUPAC"] += 1
            if nt in base_types:
                nt_counts["ACGT"] += 1
            elif nt in amb_types:
                nt_counts["amb"] += 1
            elif nt in gap_types:
                nt_counts["gap"] += 1
            
                

        return(nt_counts)
        


    @staticmethod
    def phred_score_to_int(phred_scores: str) -> list:
        int_scores = []
        for score in phred_scores:
            int_scores.append(ord(score)-33)
        return(int_scores)
        


class ProteinSeq(BioSeq):
    def __init__(self, sequence_name: str, sequence: str) -> None:
        self.name = sequence_name
        self.sequence = sequence.upper()
    

    def __repr__(self):
        return f"< Protein sequence object {self.name} of length {len(self)} AA >"
    
    def __add__(self, other):
        return(ProteinSeq(self.name, self.sequence+other.sequence))


    @classmethod
    def from_fasta(cls, input_file: Path):
        startflag = True
        with open(input_file) as f:
            for line in f:
                line = line.rstrip('\n')
                if line[0] == ">":
                    if startflag:
                        seq = ""
                        startflag = False
                    else:
                        f.close()
                        print(f"Fasta file {input_file} contains more than one entry. Only first entry is loaded into DnaSeq object. To load all entries use Fasta.from_file()\n", file = sys.stderr)
                        return(cls(sequence_name, seq))
                    sequence_name = line[1:]

                else:
                    seq += line
        f.close()
        return cls(sequence_name, seq)
    

    def print_fasta(self, output_file: str, linelength: int = None) -> None:
        printlines = []
        if linelength is None:
            printlines += [">"+self.name,self.sequence]
        else:
            printlines += [">"+self.name] +  [(self.sequence[i:i+linelength]) for i in range(0, len(self.sequence), linelength)]
        
        o = open(output_file,"w")
        o.write("\n".join(printlines)+"\n")
        o.close
        return None


class Fasta:

    def __init__(self, bio_seqs: list[BioSeq]) -> None:
        self.entries = bio_seqs

        ### Indexing for entries to allow for easy retrieval by sequence name, i.e. Fasta["sequence_name"] = [DnaSeq object]
        self.name_index = {}
        for i, bio_seq in enumerate(self.entries):
            if bio_seq.name in self.name_index:
                self.name_index[bio_seq.name].append(i)
            else:
                self.name_index[bio_seq.name] = [i]
    
    # Alternative constructor that initiates from a fasta file
    @classmethod
    def from_file(cls, input_file: str):
        entries = "Start"
        skipped_lines_count = 0
        with open(input_file) as f:
            for line in f:
                line = line.rstrip('\n')
                if not len(line) == 0:
                    if line[0] == ">":
                        if entries == "Start":
                            entries = []
                        else:
                            entries.append(BioSeq(sequence_name,new_seq))
                        new_seq = ""
                        sequence_name = line[1:]

                    else:
                        new_seq += line
                else:
                    skipped_lines_count += 1
        entries.append(BioSeq(sequence_name,new_seq))
        f.close()
        if skipped_lines_count > 0:
            print(f"Warning: {skipped_lines_count} blank lines were skipped when loading fasta file")
        return cls(entries)
    
    def check_sequence_type(self):
        
        return

    def __iter__(self):
        for entry in self.entries:
            yield entry
    
    def items(self) -> tuple[str, str]:
        for entry in self.entries:
            yield entry.name, entry.sequence

    def iterseqs(self) -> str:
        for entry in self.entries:
            yield entry.sequence
    
    def __getitem__(self,name) -> BioSeq:
        return [ self.entries[i] for i in self.name_index[name] ]
    
    # len returns total number of nucleotides in fasta file
    def __len__(self):
        length = 0
        for sequence in self.iterseqs():
            length += len(sequence)
        return(length)
    
    def __repr__(self):
        return f"< Fasta object containing {len(self.entries)} sequences, total length {len(self)} bp >"


            
    def write(self, output_file: str = None, linelength: int = None) -> None:
        printlines = []
        if linelength is None:
            for name, sequence in self.items():
                printlines += [">"+name,sequence]
        else:
            for name, sequence in self.items():
                printlines += [">"+name] +  [(sequence[i:i+linelength]) for i in range(0, len(sequence), linelength)]
        if output_file is not None:
            o = open(output_file,"w")
            o.write("\n".join(printlines)+"\n")
            o.close
        else:
            for line in printlines:
                print(line)
        return None

    # Append a new BioSeq object to exisisting Fasta object
    def append(self, entry: BioSeq) -> None:
        self.entries.append(entry)
        if entry.name in self.name_index:
            self.name_index[entry.name].append(len(self.entries)-1)
        else:
            self.name_index[entry.name] = [len(self.entries)-1]
    
    ### Add all entries from another Fasta object to entries in self
    def add(self, other: BioSeq):
        for entry in other:
            self.append(entry)
        return None

    def concat_seq(self):
        combined_seq = ""
        for name, sequence in self.items():
            combined_seq += sequence
        return(combined_seq)


    @property
    def sequence_names(self):
        seq_names = []
        for entry in self.entries:
            seq_names.append(entry.name)
        return seq_names
    
    @property
    def sequences(self) -> BioSeq:
        seqs = []
        for entry in self.entries:
            seqs.append(entry.sequence)
        return seqs

    def uniquify(self) -> BioSeq:
        seq_to_names = {}
        for name, sequence in self.items():
            if sequence in seq_to_names:
                seq_to_names[sequence].append(name)
            else:
                seq_to_names[sequence] = [name]
        ordered_seqs = [k for k, v in sorted(seq_to_names.items(), key=lambda item: len(item[1]))]
        count_str_length = len(str(len(ordered_seqs)))
        seq_count = 0
        seq_to_unique_name = {}
        unique_fasta = Fasta([]) ### initialize empty Fasta object
        for sequence in ordered_seqs:
            seq_count += 1
            seq_count_str = "0"*(count_str_length-len(str(seq_count)))+str(seq_count)
            seq_unique_name = f"unique_sequence_{seq_count_str}"
            unique_fasta.append(BioSeq(name=seq_unique_name,sequence=sequence))
            seq_to_unique_name[sequence] = seq_unique_name
        
        seq_info = []
        for name, sequence in self.items():
            seq_info.append( (name, seq_to_unique_name[sequence], len(seq_to_names[sequence]), sequence) )
        
        unique_fasta.uniquify_info = seq_info
        return unique_fasta


class NucleotideFasta(Fasta):


    def __init__(self, dna_seqs: list[DnaSeq]) -> None:
        self.entries = dna_seqs

        ### Indexing for entries to allow for easy retrieval by sequence name, i.e. Fasta["sequence_name"] = [DnaSeq object]
        self.name_index = {}
        for i, dna_seq in enumerate(self.entries):
            if dna_seq.name in self.name_index:
                self.name_index[dna_seq.name].append(i)
            else:
                self.name_index[dna_seq.name] = [i]
    
    # Alternative constructor that initiates from a fasta file
    @classmethod
    def from_file(cls, input_file: str):
        entries = "Start"
        skipped_lines_count = 0
        with open(input_file) as f:
            for line in f:
                line = line.rstrip('\n')
                if not len(line) == 0:
                    if line[0] == ">":
                        if entries == "Start":
                            entries = []
                        else:
                            entries.append(DnaSeq(sequence_name,new_seq))
                        new_seq = ""
                        sequence_name = line[1:]

                    else:
                        new_seq += line
                else:
                    skipped_lines_count += 1
        entries.append(DnaSeq(sequence_name,new_seq))
        f.close()
        if skipped_lines_count > 0:
            print(f"Warning: {skipped_lines_count} blank lines were skipped when loading fasta file")
        return cls(entries)
    
    def __repr__(self):
        return f"< NucleotideFasta object containing {len(self.entries)} sequences, total length {len(self)} bp >"
        
    def uniquify(self) -> DnaSeq:
        seq_to_names = {}
        for name, sequence in self.items():
            if sequence in seq_to_names:
                seq_to_names[sequence].append(name)
            else:
                seq_to_names[sequence] = [name]
        ordered_seqs = [k for k, v in sorted(seq_to_names.items(), key=lambda item: len(item[1]))]
        count_str_length = len(str(len(ordered_seqs)))
        seq_count = 0
        seq_to_unique_name = {}
        unique_fasta = Fasta([]) ### initialize empty Fasta object
        for sequence in ordered_seqs:
            seq_count += 1
            seq_count_str = "0"*(count_str_length-len(str(seq_count)))+str(seq_count)
            seq_unique_name = f"unique_sequence_{seq_count_str}"
            unique_fasta.append(DnaSeq(name=seq_unique_name,sequence=sequence))
            seq_to_unique_name[sequence] = seq_unique_name
        
        seq_info = []
        for name, sequence in self.items():
            seq_info.append( (name, seq_to_unique_name[sequence], len(seq_to_names[sequence]), sequence) )
        
        unique_fasta.uniquify_info = seq_info
        return unique_fasta

    def translate(self):
        protein_fasta = ProteinFasta([])
        for entry in self:
            protein_fasta.append(entry.translate())
        return protein_fasta


    @property
    def iupac_counts(self):
        combined_seq = self.concat_seq()
        base_types = ["A","C","G","T"]
        amb_types = ["R","Y","S","W","K","M","B","D","H","M"]
        gap_types = [".","-"]
        nt_counts = {"A": 0,"C": 0,"G": 0,"T": 0, "N": 0, 
                     "R": 0,"Y": 0,"S": 0,"W": 0,"K": 0,"M": 0,"B": 0,"D": 0,"H": 0,"M": 0,
                     ".": 0,"-": 0,
                     "ACGT": 0, "amb": 0, "N": 0, "gap": 0, "nonIUPAC": 0}
        for nt in combined_seq:
            try:
                nt_counts[nt] += 1
            except KeyError:
                nt_counts["nonIUPAC"] += 1
            if nt in base_types:
                nt_counts["ACGT"] += 1
            elif nt in amb_types:
                nt_counts["amb"] += 1
            elif nt in gap_types:
                nt_counts["gap"] += 1

        return(nt_counts)
    
    def batch_iupac_counts(input_directory):
        input_directory = Path(input_directory)
        iupac_counts_dict = {}
        for p in input_directory.iterdir():
            if p.name.endswith( (".fna", ".fa", ".fasta") ):
                fasta = NucleotideFasta.from_file(p)
                iupac_counts_dict[p.name] = fasta.iupac_counts
        return(iupac_counts_dict)



class ProteinFasta(Fasta):

    def __init__(self, protein_seqs: list[ProteinSeq]) -> None:
        self.entries = protein_seqs

        ### Indexing for entries to allow for easy retrieval by sequence name, i.e. ProteinFasta["sequence_name"] = ProteinSeq object
        self.name_index = {}
        for i, seq in enumerate(self.entries):
            self.name_index[seq.name] = i
    
    # Alternative constructor that initiates from a fasta file
    @classmethod
    def from_file(cls, input_file: str):
        entries = "Start"
        skipped_lines_count = 0
        with open(input_file) as f:
            for line in f:
                line = line.rstrip('\n')
                if not len(line) == 0:
                    if line[0] == ">":
                        if entries == "Start":
                            entries = []
                        else:
                            entries.append(ProteinSeq(sequence_name,new_seq))
                        new_seq = ""
                        sequence_name = line[1:]

                    else:
                        new_seq += line
                else:
                    skipped_lines_count += 1
        entries.append(ProteinSeq(sequence_name,new_seq))
        f.close()
        if skipped_lines_count > 0:
            print(f"Warning: {skipped_lines_count} blank lines were skipped when loading fasta file")
        return cls(entries)

    
    # Alternative constructor that initiates from a fasta file
    @classmethod
    def from_file(cls, input_file: str):
        entries = "Start"
        with open(input_file) as f:
            for line in f:
                line = line.rstrip('\n')
                if line[0] == ">":
                    if entries == "Start":
                        entries = []
                    else:
                        entries.append(ProteinSeq(sequence_name,new_seq))
                    new_seq = ""
                    sequence_name = line[1:]

                else:
                    new_seq += line
        entries.append(ProteinSeq(sequence_name,new_seq))
        f.close()
        return cls(entries)
    
    def __repr__(self):
        return f"< ProteinFasta object containing {len(self.entries)} sequences, total length {len(self)} bp >"

    
    def uniquify(self) -> ProteinSeq:
        seq_to_names = {}
        for name, sequence in self.items():
            if sequence in seq_to_names:
                seq_to_names[sequence].append(name)
            else:
                seq_to_names[sequence] = [name]
        ordered_seqs = [k for k, v in sorted(seq_to_names.items(), key=lambda item: len(item[1]))]
        count_str_length = len(str(len(ordered_seqs)))
        seq_count = 0
        seq_to_unique_name = {}
        unique_fasta = ProteinFasta([]) ### initialize empty Fasta object
        for sequence in ordered_seqs:
            seq_count += 1
            seq_count_str = "0"*(count_str_length-len(str(seq_count)))+str(seq_count)
            seq_unique_name = f"unique_sequence_{seq_count_str}"
            unique_fasta.append(ProteinSeq(name=seq_unique_name,sequence=sequence))
            seq_to_unique_name[sequence] = seq_unique_name
        
        seq_info = []
        for name, sequence in self.items():
            seq_info.append( (name, seq_to_unique_name[sequence], len(seq_to_names[sequence]), sequence) )
        
        unique_fasta.uniquify_info = seq_info
        return unique_fasta


    def __getitem__(self,name) -> ProteinSeq:
        return [ self.entries[i] for i in self.name_index[name] ]
    
def translate_dna(dna_string: str):
    seq = dna_string.replace("-","")
    table = { 
        'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M', 
        'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T', 
        'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K', 
        'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R', 
        'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L', 
        'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P', 
        'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q', 
        'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R', 
        'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V', 
        'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A', 
        'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E', 
        'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G', 
        'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S', 
        'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L', 
        'TAC':'Y', 'TAT':'Y', 'TAA':'*', 'TAG':'*', 
        'TGC':'C', 'TGT':'C', 'TGA':'*', 'TGG':'W', 
    } 
    protein_seq = "" 
    mod = len(seq)%3
    if not mod == 0:
        print(f"Warning: number of nucleotides in sequence is not divisible by 3. Ignoring {mod} nt at the end of sequence when translating to protein.", file=sys.stderr)
    illegal_codons = 0
    for i in range(0, len(seq)-mod, 3):
        codon = seq[i:i + 3]
        if len(set(codon)-{"A","C","G","T"}) > 0:
            illegal_codons += 1
            protein_seq += "X"
        else:
            protein_seq+= table[codon]
    if illegal_codons > 0:
        print(f"WARNING: {illegal_codons} codons containing non A/C/G/T found. X has been inserted on those positions.", file=sys.stderr)
    return(protein_seq)



class WgsData:

    def __init__(self, file_paths: dict[dict], sample_data: dict[dict] = None):
        """
        Default initialization is from file paths provided in a dictionary with sample name as keys, like this:
        file_paths =
        {<sample_name>: {
            "assembly_file": <assembly_file_path>,
            "r1_file": <r1_file_path>,
            "r2_file": <r2_file_path>
            }
        }
        and metadata in a similar format with sample names as keys and a dictionary of attribute: value pairs as values.
        {<sample_name>: {
            "serotype": "a",
            "MLST": "15",
            "source": blood,
            ....
            }
        }



        To initialize from folders without using a metadata sheet, use WgsData.from_folders(<assembly_data_folder>, <read_data_folder>)
        Both assembly_data_folder and read_data_folder can be set to None if you have only reads or only assemblies.
        """
        self.sample_names = file_paths.keys()
        self.file_paths = file_paths
        self.paired_end_read_files = []
        self.assembly_files = []
        for files in file_paths.values():
            if "r1_file" in files and "r2_file" in files and files["r1_file"] is not None and files["r2_file"]:
                self.paired_end_read_files.append((files["r1_file"], files["r2_file"]))
            else:
                self.paired_end_read_files.append((None, None))
            if "assembly_file" in files and files["assembly_file"] is not None:
                self.assembly_files.append(files["assembly_file"])
            else:
                self.assembly_files.append(None)
        self.sample_data = sample_data
        if sample_data:
            self.sample_data_df = pandas.DataFrame.from_dict(
                sample_data, orient="index"
            )
            self.sample_data_df.index = self.sample_names
            self.sample_data_columns = list(self.sample_data_df.columns.values)
        else:
            self.sample_data_columns = None
            self.sample_data_df = None

    @classmethod
    def from_samplesheet(cls, sample_sheet: Path, base_folder: Path = None):
        """
        Load data from samplesheet into dictionary.
        Data sheet must contain column names "sample_name", "assembly_file", "r1_file", "r2_file" (if files exist)
        If there are duplicates in "sample_name" column, entries after the first instance will have added _2, _3, _4 etc to their name
        """

        df = pandas.read_csv(sample_sheet, sep="\t")
        file_paths_df = df[["sample_name", "assembly_file", "r1_file", "r2_file"]]
        file_paths = {}
        duplicate_sample_count = 0
        ignored_assembly_count = 0
        ignored_r1_count = 0
        ignored_r2_count = 0
        uniq_sample_names = []
        for index, row in file_paths_df.iterrows():
            sample_name = row["sample_name"]
            if not sample_name in uniq_sample_names:
                file_paths[sample_name] = row.to_dict()
                uniq_sample_names.append(sample_name)
            else:
                name_counter = 1
                sample_name_uniq = f"{sample_name}_{name_counter}"
                while sample_name_uniq in uniq_sample_names:
                    name_counter += 1
                    sample_name_uniq = f"{sample_name}_{name_counter}"
                file_paths[sample_name_uniq] = row.to_dict()
                uniq_sample_names.append(sample_name_uniq)
                duplicate_sample_count += 1
        if base_folder:
            base_folder = Path(base_folder).absolute()
            for files in file_paths.values():
                files["assembly_file"] = base_folder.joinpath(files["assembly_file"])
                files["r1_file"] = base_folder.joinpath(files["r1_file"])
                files["r2_file"] = base_folder.joinpath(files["r2_file"])
        else:
            base_folder = Path(sample_sheet).parent
            for sample, files in file_paths.items():
                if files["assembly_file"] is not None:
                    try:
                        files["assembly_file"] = base_folder.joinpath(
                            files["assembly_file"]
                        ).resolve()
                    except TypeError as e:
                        ignored_assembly_count += 1
                        files["assembly_file"] = None
                if files["r1_file"] is not None:
                    try:
                        files["r1_file"] = base_folder.joinpath(
                            files["r1_file"]
                        ).resolve()
                    except TypeError as e:
                        ignored_r1_count += 1
                        files["r1_file"] = None
                if files["r2_file"] is not None:
                    try:
                        files["r2_file"] = base_folder.joinpath(
                            files["r2_file"]
                        ).resolve()
                    except TypeError as e:
                        ignored_r2_count += 1
                        files["r2_file"] = None
        if ignored_assembly_count > 0:
            print(f"{ignored_assembly_count} assembly file paths ignored")
        if ignored_r1_count > 0:
            print(f"{ignored_r1_count} assembly file paths ignored")
        if ignored_assembly_count > 0:
            print(f"{ignored_r2_count} assembly file paths ignored")
        df["sample_name_uniq"] = uniq_sample_names
        df.set_index("sample_name_uniq", inplace=True)
        sample_data = df.to_dict("index")
        return cls(file_paths=file_paths, sample_data=sample_data)

    @classmethod
    def from_folders(
        cls,
        assembly_data_folder: Path = None,
        paired_end_read_data_folder: Path = None,
        fasta_file_pattern=r"(?P<sample_name>.+?)(\.fa|\.fna|\.fasta)$",
        paired_end_reads_pattern=r"(?P<sample_name>.+?)(?P<sample_number>(_S[0-9]+)?)(?P<lane>(_L[0-9]+)?)[\._]R?(?P<paired_read_number>[1|2])(?P<set_number>(_[0-9]+)?)(?P<file_extension>\.fastq\.gz)",
    ):
        if assembly_data_folder is not None:
            assembly_data_dict = cls.parse_folder_for_fasta_files(
                data_folder=assembly_data_folder, fasta_file_pattern=fasta_file_pattern
            )
        if assembly_data_folder is not None and len(assembly_data_dict) > 0:
            assembly_df = pandas.DataFrame.from_dict(assembly_data_dict, orient="index")
            assembly_df.columns = ["assembly_file"]
        else:
            assembly_df = pandas.DataFrame(
                {
                    "assembly_file": [],
                }
            )
        if paired_end_read_data_folder is not None:
            paired_end_read_data_dict, unmatched_read_files = (
                cls.parse_folder_for_paired_end_reads(
                    data_folder=paired_end_read_data_folder,
                    paired_end_reads_pattern=paired_end_reads_pattern,
                )
            )
        if paired_end_read_data_folder is not None and len(paired_end_read_data_dict) > 0:
            paired_end_read_df = pandas.DataFrame(paired_end_read_data_dict).transpose()
        else:
            paired_end_read_df = pandas.DataFrame(
                {"sample_name": [], "r1_file": [], "r2_file": []}
            )

        combined_df = assembly_df.merge(
            paired_end_read_df, how="outer", left_index=True, right_on="sample_name"
        )
        combined_df.replace(numpy.nan, None, inplace=True)
        combined_df = combined_df[
            ["sample_name", "assembly_file", "r1_file", "r2_file"]
        ]
        combined_dict = combined_df.to_dict("index")

        return cls(file_paths=combined_dict)

    def __iter__(self):
        if self.sample_data is None:
            for sample_name, files in self.file_paths.items():
                yield sample_name, files, None
        else:
            for sample_name, files in self.file_paths.items():
                yield sample_name, files, self.sample_data[sample_name]
    
    def __len__(self):
        return len(self.sample_names)

    def get_missing_files(self):
        missing_assembly_files = {}
        missing_read_files = {}
        for sample_name, assembly_file, r1_file, r2_file in self.filepaths():
            if assembly_file is None or not assembly_file.exists():
                missing_assembly_files[sample_name] = assembly_file
            if (
                r1_file is None
                or r2_file is None
                or not r1_file.exists()
                or not r2_file.exists()
            ):
                missing_read_files[sample_name] = (r1_file, r2_file)
        return missing_assembly_files, missing_read_files

    def wgs_files_exist(self):
        missing_assembly_files, missing_read_files = self.get_missing_files()
        if len(missing_assembly_files) == 0 and len(missing_read_files) == 0:
            return True
        else:
            return False

    @classmethod
    def from_dataframe(cls, pandas_dataframe: pandas.DataFrame):
        sample_data = pandas_dataframe.to_dict("index")
        sample_data.replace(numpy.nan, None, inplace=True)
        return cls(file_paths=sample_data)

    def filepaths(self):
        for i, sample_name in enumerate(self.sample_names):
            yield sample_name, self.assembly_files[i], self.paired_end_read_files[i][
                0
            ], self.paired_end_read_files[i][1]

    def readpaths(self):
        for i, sample_name in enumerate(self.sample_names):
            yield sample_name, self.paired_end_read_files[0], self.paired_end_read_files[1]

    def assemblypaths(self):
        for i, sample_name in enumerate(self.sample_names):
            yield sample_name, self.assembly_files[i]

    def __repr__(self):
        if self.sample_data:
            return f"< WgsData object with {len(self.sample_names)} samples and {len(self.sample_data_columns)} metadata columns >"
        else:
            return f"< WgsData object with data from {len(self.sample_names)} samples >"

    @staticmethod
    def parse_folder_for_paired_end_reads(
        data_folder: Path,
        paired_end_reads_pattern: str = r"(?P<sample_name>.+?)(?P<sample_number>(_S[0-9]+)?)(?P<lane>(_L[0-9]+)?)[\._]R?(?P<paired_read_number>[1|2])(?P<set_number>(_[0-9]+)?)(?P<file_extension>\.fastq\.gz)",
    ) -> tuple[dict[dict], set]:
        """
        Parses a provided folder for paired end reads matching the provided regex pattern
        Returns a dictionary with sample names as keys and a dict with keys "r1_file" and "r2_file" and filenames as values
        In case of duplicate sample names, the <sample_name>_<sample_number> combination will be used instead
        If <sample_name>_<sample_number> is also not unique, the key used for sample name will simply be the name of the r1_file
        """
        sample_list = []
        unmatched_files = set()
        sample_names = set()
        duplicate_sample_names = set()
        duplicate_sample_names_uniq = set()
        data_folder = Path(data_folder).resolve()
        for f in data_folder.iterdir():
            fname = f.name
            fname_match = re.match(paired_end_reads_pattern, fname)
            if fname_match:
                sample_name = fname_match.group("sample_name")
                if fname_match.group("paired_read_number") is not None:
                    read_number = fname_match.group("paired_read_number")
                    if fname_match.group("sample_number") is not None:
                        sample_name_uniq = (
                            f"{sample_name}{fname_match.group('sample_number')}"
                        )
                    else:
                        sample_name_uniq = f"{sample_name}"
                    if read_number == "1":
                        r1_file = f.absolute()
                        r2_file = data_folder.joinpath(
                            fname[: fname_match.start("paired_read_number")]
                            + "2"
                            + fname[fname_match.end("paired_read_number") :]
                        )
                        if r2_file.exists():
                            sample_list.append(
                                (sample_name, sample_name_uniq, r1_file, r2_file)
                            )
                            if sample_name in sample_names:
                                duplicate_sample_names.add(sample_name)
                                if sample_name_uniq in duplicate_sample_names:
                                    duplicate_sample_names_uniq.add(sample_name_uniq)
                            else:
                                sample_names.add(sample_name)
                        else:
                            unmatched_files.add(fname)
                    if read_number == "2":
                        r1_file = data_folder.joinpath(
                            fname[: fname_match.start("paired_read_number")]
                            + "1"
                            + fname[fname_match.end("paired_read_number") :]
                        )
                        if not r1_file.exists():
                            unmatched_files.add(fname)
        sample_file_dict = {}
        for sample_name, sample_name_uniq, r1_file, r2_file in sample_list:
            if sample_name in duplicate_sample_names:
                if sample_name in duplicate_sample_names_uniq:
                    sample_file_dict[r1_file] = {"r1_file": r1_file, "r2_file": r2_file}
                else:
                    sample_file_dict[sample_name_uniq] = {
                        "sample_name": sample_name,
                        "r1_file": r1_file,
                        "r2_file": r2_file,
                    }
            else:
                sample_file_dict[sample_name] = {
                    "sample_name": sample_name,
                    "r1_file": r1_file,
                    "r2_file": r2_file,
                }

        return sample_file_dict, unmatched_files

    @staticmethod
    def parse_folder_for_fasta_files(
        data_folder: Path,
        fasta_file_pattern: str = r"(?P<sample_name>.+?)(\.fa|\.fna|\.fasta)$",
    ) -> tuple[dict[dict], set]:
        """
        Parses a provided folder for paired end reads matching the provided regex pattern
        Returns a dictionary with sample names as keys and a dict with keys "r1_file" and "r2_file" and filenames as values
        In case of duplicate sample names, the <sample_name>_<sample_number> combination will be used instead
        If <sample_name>_<sample_number> is also not unique, the key used for sample name will simply be the name of the r1_file
        """
        sample_file_dict = {}
        data_folder = Path(data_folder).resolve()
        for f in data_folder.iterdir():
            fname = f.name
            fname_match = re.match(fasta_file_pattern, fname)
            if fname_match:
                sample_name = fname_match.group("sample_name")
                sample_file_dict[sample_name] = f
        return sample_file_dict